@@ -684,6 +684,7 @@ RenameTypeInternal(Oid typeOid, const char *newTypeName, Oid typeNamespace)
 	HeapTuple	tuple;
 	Form_pg_type typ;
 	Oid			arrayOid;
+	Oid			oldTypeOid;
 
 	pg_type_desc = heap_open(TypeRelationId, RowExclusiveLock);
 
@@ -697,13 +698,28 @@ RenameTypeInternal(Oid typeOid, const char *newTypeName, Oid typeNamespace)
 
 	arrayOid = typ->typarray;
 
-	/* Just to give a more friendly error than unique-index violation */
-	if (SearchSysCacheExists2(TYPENAMENSP,
-							  CStringGetDatum(newTypeName),
-							  ObjectIdGetDatum(typeNamespace)))
-		ereport(ERROR,
-				(errcode(ERRCODE_DUPLICATE_OBJECT),
-				 errmsg("type \"%s\" already exists", newTypeName)));
+	/* Check for a conflicting type name. */
+	oldTypeOid = GetSysCacheOid2(TYPENAMENSP,
+								 CStringGetDatum(newTypeName),
+								 ObjectIdGetDatum(typeNamespace));
+
+	/*
+	 * If there is one, see if it's an autogenerated array type, and if so
+	 * rename it out of the way.  (But we must skip that for a shell type
+	 * because moveArrayTypeName will do the wrong thing in that case.)
+	 * Otherwise, we can at least give a more friendly error than unique-index
+	 * violation.
+	 */
+	if (OidIsValid(oldTypeOid))
+	{
+		if (get_typisdefined(oldTypeOid) &&
+			moveArrayTypeName(oldTypeOid, newTypeName, typeNamespace))
+			 /* successfully dodged the problem */ ;
+		else
+			ereport(ERROR,
+					(errcode(ERRCODE_DUPLICATE_OBJECT),
+					 errmsg("type \"%s\" already exists", newTypeName)));
+	}
 
 	/* OK, do the rename --- tuple is a copy, so OK to scribble on it */
 	namestrcpy(&(typ->typname), newTypeName);
@@ -716,8 +732,12 @@ RenameTypeInternal(Oid typeOid, const char *newTypeName, Oid typeNamespace)
 	heap_freetuple(tuple);
 	heap_close(pg_type_desc, RowExclusiveLock);
 
-	/* If the type has an array type, recurse to handle that */
-	if (OidIsValid(arrayOid))
+	/*
+	 * If the type has an array type, recurse to handle that.  But we don't
+	 * need to do anything more if we already renamed that array type above
+	 * (which would happen when, eg, renaming "foo" to "_foo").
+	 */
+	if (OidIsValid(arrayOid) && arrayOid != oldTypeOid)
 	{
 		char	   *arrname = makeArrayTypeName(newTypeName, typeNamespace);
 
